import keras #for one-hot-encoding
import math 

class InMemoryImgHandler:
    """ Handles image data available in RAM

    + Splits and stores data in-memory
    + Creates one-hot-encoded targets if specified

    Attributes
    ___________
    - x_train  : np.ndarray
    - x_test   : np.ndarray
    - y_train  : np.ndarray
    - y_test   : np.ndarray

    Methods
    _______ 
    - train_val_test_split(X: np.ndarray, y: np.ndarray, ratio: tuple)
    - one_hot_encode(y: np.ndarray, num_classes: int)
    """

    
    def __init__(self, X, y, y_info, ratio=(60,20,20)):
        """
        - X: np.ndarray
            + np.nd array (n_samples, w, h, c)
        - y: np.ndarray
            + Either one-hot-encoded (n_smaple, n_classes) bin matrix
            + or numeric (n_smaple, 1) vec
        - y_info: tuple
            - y_info[0]: int 
                + num_classes
            - y_info[1]: bool 
                + True if one-hot-encoded already
        - ratio: tuple = (60,20,20)
            + (train_ratio, val_ratio, test_ratio) between 0-100 and sum to 100
        """

        if y_info[1] is False:
            y = self.one_hot_encode(y, y_info[0])
        
        # split and store                
        (self.x_train, self.y_train), (self.x_val, self.y_val), (self.x_test, self.y_test) = \
            InMemoryImgHandler.train_val_test_split(X, y, ratio)

    # ============================================================================================
    # BEG: static methods 
    # ============================================================================================
    @staticmethod
    def train_val_test_split(X, y, ratio):
        """ Splits data into train test and validation

        Parameters
        __________
        - X: np.ndarray 
            + 4-D (n_samples, w, h, c) 
        - y: np.ndarray
            + one-hot-encoded bin matrix
        - ratio: tuple 
            + (train_ratio, val_ratio, test_ratio)

        Returns 
        _______
        tuple 
            + (X_tr, y_tr), (X_val, y_val), (X_te, y_te)

        * Note: Not shufflind cz, taken care by `model.fit`
        """

        if (ratio[0] + ratio[1] + ratio[2]) != 100:
            raise Exception('Split percents\' sum must be 100!')

        n_samples = X.shape[0]
        n_train   =  math.floor( (ratio[0]/100) * n_samples )
        n_val     =  math.floor( (ratio[1]/100) * n_samples )
        n_test    =  math.floor( (ratio[2]/100) * n_samples )
        
        train_l, train_r   =             0, n_train
        val_l, val_r       =       n_train, n_train+n_val
        test_l, test_r     = n_train+n_val, n_train+n_val+n_test
        
        return (
            ( X[train_l:train_r], y[train_l:train_r] ), 
            ( X[val_l:val_r]    , y[val_l:val_r] ), 
            ( X[test_l:test_r]  , y[test_l:test_r] ), 
        )
    
    @staticmethod
    def one_hot_encode(y, num_classes):
        """ y (is n_samples, 1) 2-D vector. Not a flat array"""
        return keras.utils.to_categorical(y, num_classes)
    # ============================================================================================
    # END: static methods 
    # ============================================================================================


import os
import keras
from nextoff.models.raw import TestModel

# =======================================================================================
# BEG: abstraction for train, test and eval of model
# =======================================================================================
def train_test_model_with(data, args, savename="bestmodel"):
    """ Build temporary model and train/test/eval/plot w/ it 
    
    Parameters
    __________
    - data: data.inmemory.InMemoryImgHandler
    - args:  must follow convention
    - savename: str
        + name of the saved model inside `BestSavedModels` dir
    """
    # initialize and build
    # --------------------
    baseline_model = TestModel(args)
    baseline_model.compile(args)
    
    # Train eval, save and plot
    # -------------------------
    if 'v' in args.__dict__: v = args.v
    else: v = 0
    hist = baseline_model.fit( 
                (data.x_train, data.y_train),
                (data.x_val, data.y_val), 
                v,
                savename, 
                args
    )
    
    # Evaluate the performance (unseen data)
    baseline_model.evaluate(data.x_test, data.y_test)
    # plot
    baseline_model.plot()

    return hist
# =======================================================================================
# END: abstraction for train, test and eval of model
# =======================================================================================


class TestClass:
    def __init__(self):
        pass

    def printHello(self):
        print('Hello')

class TestClass2:
    def __init__(self):
        pass

    def printHello(self):
        print('Hello')


# Example code
if __name__ == '__main__':
    print("This should not be printed!!")


from distutils.core import setup
setup(
  name = 'nextoff',         
  packages = [
    'nextoff', 
    'nextoff.test', 
    'nextoff.data', 
    'nextoff.models',
    'nextoff.train', 
    ],  # ! Important
  version = '0.1.1.7', # ! Always update
  license='MIT',        
  description = 'A package to experiment and tune keras hyperparameters', 
  author = 'Asapanna Rakesh',                   
  author_email = 'rakeshark22@gmail.com',      
  url = 'https://github.com/rakesh4real/nextoff', 
  download_url = 'https://github.com/rakesh4real/nextoff/archive/v0.1.1.7.tar.gz',
  keywords = ['Keras', 'hyperparameters', 'hyperparameters tuning'],   
  install_requires=[        
          'keras',
          'matplotlib'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)


import matplotlib.pyplot as plt # plotting
import os # for file/folder creations
import keras
from keras.models import Model, Sequential
from keras.layers import Dense, Conv2D, MaxPooling2D, Dropout, Activation, Flatten, BatchNormalization, Input, Dense
from keras.callbacks import ModelCheckpoint
from keras.preprocessing.image import ImageDataGenerator

# =======================================================================================
# BEG: build custom model based on args
# =======================================================================================
class TestModel:
    """ Tensorflow model created using args

    Attributes
    __________
    - args: all configs

    Methods
    _______
    - build_model(args)
    - compile(args)
    - fit(train_data, val_data, v, savename, args)
       - train_data: tuple of X and y
       - val_data: tuple of X and y
    - evaluate(x_test, y_test)
       + evaluates on completely unseen data
    - save_to(save_path)
    - summary
    - plot
    """
    def __init__(self, args):
        # build model
        self.args = args
        self.model = self.build_model(args)    
    
    def build_model(self, args):
        # define input dims
        input_dims = Input(shape=args.input_shape)#(32,32,3)) # flexible batch size
        
        x = self.convs_from_seq(args.conv_seq, input_dims)
        x = Flatten()(x)
        x = self.fcs_from_seq(args.fcs_seq, x)
        
        return Model(inputs=input_dims, outputs=x)
    
    def compile(self, args):

        # todo: add other optimizers
        if   'RMSprop' in args.__dict__ : opt = keras.optimizers.RMSprop(**args.RMSprop)
        elif 'SGD'     in args.__dict__ : opt = keras.optimizers.SGD(**args.SGD)
        elif 'Adam'    in args.__dict__ : opt = keras.optimizers.Adam(**args.Adam)
        else: 
            raise Exception('Define optimizers name exactly as in keras.optimizers')

        self.model.compile(
            loss       = args.loss,
            optimizer  = opt,
            metrics    = args.metrics,
        )
    
    # ----------------------------------------------------------
    # Training and evaluation
    # ----------------------------------------------------------
    def fit(self, train_data, val_data, v, savename, args):
        """train_data, val_data are tuples of X and y"""
        if v==0:
            print("Wait for the big picture ☕️\n\nTrainng started ...")
        num_samples = train_data[0].shape[0] #X
        num_batches = num_samples / args.batch_size

        callbacks_list = []
        # custom callbacks (single only)
        if 'CustomCallback' in args.__dict__:
            callbacks_list.append(args.CustomCallback(num_batches))
        
        # checkpointer callback
        os.makedirs(f'BestSavedModels', exist_ok=True)
        check = ModelCheckpoint(
            filepath=f"./BestSavedModels/{savename}.h5",
            verbose=v, save_best_only=True
        )
        callbacks_list.append(check)
        
        # data generator
        traingen = ImageDataGenerator(**args.train_data_gen)
        valgen   = ImageDataGenerator(**args.val_data_gen)

        self.history = self.model.fit(
            traingen.flow(train_data[0], train_data[1], args.batch_size),
            validation_data = valgen.flow(val_data[0], val_data[1], batch_size=args.batch_size),
            steps_per_epoch = len(train_data[0]) / args.batch_size, 
            epochs          = args.epochs,
            shuffle         = args.shuffle,
            verbose         = v,
            callbacks       = callbacks_list
        )

        return self.history
    
    # evaluation
    def evaluate(self, x_test, y_test):
        print("Evaluating on test data...\n")
        testgen  = ImageDataGenerator(**self.args.test_data_gen)

        test_loss, test_acc = \
            self.model.evaluate(
                testgen.flow(x_test, y_test, batch_size=self.args.batch_size)
            )
        print(f"+ Test Loss\t:{test_loss}\n+ Test Acc\t:{test_acc}")
    
    # save
    def save_to(self, save_path):
        self.model.save(save_path)
    
    # -----------------------------------------------------------
    # plot
    # -----------------------------------------------------------
    def plot(self):
        history_dict = self.history.history
        
        # Plot loss chart
        loss_values     = history_dict['loss']
        val_loss_values = history_dict['val_loss']
        epochs          = range(1, len(loss_values) + 1)
        
        line1 = plt.plot(epochs, val_loss_values, label='Validation/Test Loss')
        line2 = plt.plot(epochs, loss_values, label='Training Loss')
        
        plt.setp(line1, linewidth=2.0, marker = '+', markersize=10.0)
        plt.setp(line2, linewidth=2.0, marker = '4', markersize=10.0)
        plt.xlabel('Epochs') 
        plt.ylabel('Loss')
        plt.grid(True)
        plt.legend()
        
        # Plot acc chart
        acc_values      = history_dict['accuracy']
        val_acc_values  = history_dict['val_accuracy']
        epochs          = range(1, len(loss_values) + 1)

        line1 = plt.plot(epochs, val_acc_values, label='Validation/Test Accuracy')
        line2 = plt.plot(epochs, acc_values, label='Training Accuracy')
        
        plt.setp(line1, linewidth=2.0, marker = '+', markersize=10.0)
        plt.setp(line2, linewidth=2.0, marker = '4', markersize=10.0)
        plt.xlabel('Epochs') 
        plt.ylabel('Accuracy')
        plt.grid(True)
        plt.legend()        
        
        # display
        plt.show()
    
    def summary(self):
        keras.utils.plot_model(
            self.model,
            to_file="model.png",
            show_shapes=False,
            #show_dtype=False,
            show_layer_names=True,
            rankdir="TB",
            expand_nested=False,
            dpi=96,
        )
        return self.model.summary()    
    # ===========================================================
    # BEG: build_model helpers
    # ===========================================================
    def convs_from_seq(self, seq, x):
        """
        + x is input which will be forwarded in series 
        + seq is of format
            [
                {"out_ch": int, "z": int, 'act': 'relu', 'bn': True, 'p': 0.5, 'L1': 1e-4},
                {"out_ch": int, "z": int, 'act': 'relu', "maxpool_z": 2},
                # -------------------------------------------------------
                {"out_ch": int, "z": int, 'act': 'relu'},
                ....
            ]
        + regularize all if `args.regularize_all = L1 / L2 / L1L2` 
        """
        # bulid series
        series = []
        for config in seq:
            
            # append convs and nonlin one-by-one                
            # regularize all
            kernel_regularizer = None
            if ('L1' in self.args.__dict__):
                kernel_regularizer =  keras.regularizers.l1(l1=self.args.__dict__['L1'])
            elif ('L2' in self.args.__dict__):
                kernel_regularizer =  keras.regularizers.l2(l2=self.args.__dict__['L2'])
            elif ('L1L2' in self.args.__dict__):
                kernel_regularizer =  keras.regularizers.l1_l2(
                    l1=self.args.__dict__['L1L2'][0], 
                    l2=self.args.__dict__['L1L2'][1]
                )
            # regularize >>>individually<<<
            # (OVERWRITE `kernel_regularizer` created by regularize all)
            if 'L1' in config:
                kernel_regularizer =  keras.regularizers.l1(l1=config['L1'])
            elif 'L2' in config:
                kernel_regularizer =  keras.regularizers.l2(l2=config['L2'])
            elif 'L1L2' in config:
                kernel_regularizer =  keras.regularizers.l1_l2(
                    l1=config['L1L2'][0],
                    l2=config['L1L2'][1],
                )
            
            series.append(
                self.__get_conv(
                    out_ch              = config['out_ch'],
                    z                   = config['z'],
                    padding             = 'same',
                    kernel_regularizer  = kernel_regularizer
                )
            )
            
            # activation function
            series.append(self.__get_act(config['act']))
            
            # pool if specified
            if 'maxpool_z' in config.keys():
                # append pool one-by-one
                series.append(self.__get_pool(config['maxpool_z']))
            
            # (before bn)
            if 'p' in config.keys():
                series.append(self.__get_dropout(config["p"]))
                
            # batchnorm (after dropout)
            if 'bn' in config.keys():
                if config['bn'] is True:
                    # append bn one-by-one
                    series.append(self.__get_bn())

        # forward through series
        """ #OVERKILL
        return Sequential(series)
        """
        for layer in series:
            x = layer(x)
        return x

    def fcs_from_seq(self, seq, x):
        """
        + x is input
        + seq is of format
            [
                {"out_nodes": int, "act": 'relu', 'bn': True},
                {"out_nodes": int, "act": 'relu', 'bn': Fale, 'p': 0.5 },
                ....
            ]
        """
        # build series
        series = []
        for config in seq:
            
            # regularize all
            kernel_regularizer = None
            if ('L1' in self.args.__dict__):
                kernel_regularizer =  keras.regularizers.l1(l1=self.args.__dict__['L1'])
            elif ('L2' in self.args.__dict__):
                kernel_regularizer =  keras.regularizers.l2(l2=self.args.__dict__['L2'])
            elif ('L1L2' in self.args.__dict__):
                kernel_regularizer =  keras.regularizers.l1_l2(
                    l1=self.args.__dict__['L1L2'][0], 
                    l2=self.args.__dict__['L1L2'][1]
                )
            # regularize >>>individually<<<
            # (OVERWRITE `kernel_regularizer` created by regularize all)
            if 'L1' in config:
                kernel_regularizer =  keras.regularizers.l1(l1=config['L1'])
            elif 'L2' in config:
                kernel_regularizer =  keras.regularizers.l2(l2=config['L2'])
            elif 'L1L2' in config:
                kernel_regularizer =  keras.regularizers.l1_l2(
                    l1=config['L1L2'][0],
                    l2=config['L1L2'][1],
                )
                
            series.append(
                self.__get_dense(
                    out_nodes           = config["out_nodes"], 
                    act                 = config["act"],
                    kernel_regularizer  = kernel_regularizer
                )
            )
            
            # batchnorm (before dropout)
            if 'bn' in config:
                if config['bn'] is True:
                    # append bn one-by-one
                    series.append(self.__get_bn())
                    
            # dropout (after batchnorm)
            if 'p' in config:
                series.append(self.__get_dropout(config["p"]))
        
        # forward through series
        """ # OVERKILL
        return Sequential(series)
        """
        for layer in series:
            x = layer(x)
        return x

        
    # helpers start ---------------------------------
    # general
    def __get_bn(self):
        return BatchNormalization()
        
    def __get_act(self, name):
        """`name` is a string. eg. 'relu'"""
        return Activation(name)
    
    # for convs_seq
    def __get_conv(self, out_ch, z, padding, kernel_regularizer):
        """
        For "SAME" padding, if you use a stride of 1, 
        the layer's outputs will have the same spatial 
        dimensions as its inputs.
        """
        return Conv2D(out_ch, (z, z), padding=padding, 
                      strides=(1, 1), kernel_regularizer=kernel_regularizer)
    
    def __get_pool(self, z):
        return MaxPooling2D(pool_size=(z, z))
    
    # for fcs_Seq
    def __get_dense(self, out_nodes, act, kernel_regularizer):
        """out_nodes is int and act is string """
        return Dense(out_nodes, activation=act, kernel_regularizer=kernel_regularizer)
    
    def __get_dropout(self, p):
        """ p if float [0, 1]"""
        return Dropout(p)
    # helpers end ------------------------------------
    # ===========================================================
    # END: build_model helpers 
    # ===========================================================
# =======================================================================================
# END: build custom model based on args
# =======================================================================================
